from typing import Any, Callable, Optional

from fastapi import params

from lanarky.utils import model_dump

from .resources import OpenAIResource
from .utils import create_request_model, create_response_model


def Depends(
    dependency: Optional[Callable[..., Any]],
    *,
    dependency_kwargs: dict[str, Any] = {},
    use_cache: bool = True
) -> params.Depends:
    """Dependency injection for OpenAI.

    Args:
        dependency: a "dependable" resource factory callable.
        dependency_kwargs: kwargs to pass to resource dependency.
        use_cache: use_cache parameter of `fastapi.Depends`.
    """
    try:
        resource = dependency()
    except TypeError:
        raise TypeError("set default values for all dependency parameters")

    if not isinstance(resource, OpenAIResource):
        raise TypeError("dependency must return a OpenAIResource instance")

    request_model = create_request_model(resource)
    response_model = create_response_model(resource)

    async def resource_dependency(
        request: request_model,
        resource: OpenAIResource = params.Depends(dependency, use_cache=use_cache),
    ) -> response_model:
        resource_kwargs = {**model_dump(request), **dependency_kwargs}

        return await resource(**resource_kwargs)

    return params.Depends(resource_dependency, use_cache=use_cache)



import json
from contextlib import contextmanager
from typing import Any, Iterator, Optional

import httpx
from httpx_sse import ServerSentEvent, connect_sse
from websockets.sync.client import connect as websocket_connect

from lanarky.websockets import DataMode


class StreamingClient:
    """Test client for streaming server-sent events."""

    def __init__(
        self,
        base_url: str = "http://localhost:8000",
        client: Optional[httpx.Client] = None,
    ):
        """Constructor method.

        Args:
            base_url: The base URL of the server.
            client: The HTTP client to use.
        """
        self.base_url = base_url
        self.client = client or httpx.Client()

    def stream_response(
        self, method: str, path: str, **kwargs: dict[str, Any]
    ) -> Iterator[ServerSentEvent]:
        """Stream data from the server.

        Args:
            method: The HTTP method to use.
            path: The path to stream from.
            **kwargs: The keyword arguments to pass to the HTTP client.
        """
        url = self.base_url + path
        with connect_sse(self.client, method, url, **kwargs) as event_source:
            for sse in event_source.iter_sse():
                yield sse


class WebSocketClient:
    """Test client for WebSockets.

    Supports 3 data modes: JSON, TEXT, and BYTES.
    """

    def __init__(
        self, uri: str = "ws://localhost:8000/ws", mode: DataMode = DataMode.JSON
    ):
        """Constructor method.

        Args:
            uri: The URI of the websocket.
            mode: The data mode to use.
        """
        self.uri = uri
        self.mode = mode
        self.websocket = None

    @contextmanager
    def connect(self):
        """Connect to a websocket and yield data from it."""
        with websocket_connect(self.uri) as websocket:
            self.websocket = websocket
            yield self
            self.websocket = None

    def send(self, message: Any):
        """Send data to the websocket.

        Args:
            message: The data to send.
        """
        if self.websocket:
            if self.mode == DataMode.JSON:
                message = json.dumps(message)
            elif self.mode == DataMode.TEXT:
                message = str(message)
            elif self.mode == DataMode.BYTES:
                message = message.encode("utf-8")
            self.websocket.send(message)

    def receive(self, mode: DataMode = None):
        """Receive data from the websocket.

        Args:
            mode: The data mode to use.
        """
        mode = mode or self.mode
        if self.websocket:
            response = self.websocket.recv()
            if mode == DataMode.JSON:
                response = json.loads(response)
            elif mode == DataMode.TEXT:
                response = str(response)
            elif mode == DataMode.BYTES:
                response = response.decode("utf-8")
            return response

    def stream_response(self):
        """Stream data from the websocket.

        Streaming stops until the websocket is closed or
        the `end` event is received.
        """
        if self.websocket:
            while True:
                response = self.receive(mode=DataMode.JSON)
                if response["event"] == "end":
                    break
                yield response



from typing import Any

from fastapi import status
from starlette.types import Send

from lanarky.adapters.openai.resources import Message
from lanarky.events import Events, ServerSentEvent, ensure_bytes
from lanarky.logging import logger
from lanarky.responses import HTTPStatusDetail
from lanarky.responses import StreamingResponse as _StreamingResponse

from .resources import OpenAIResource


class StreamingResponse(_StreamingResponse):
    """StreamingResponse class for OpenAI resources."""

    def __init__(
        self,
        resource: OpenAIResource,
        messages: list[Message],
        *args: Any,
        **kwargs: dict[str, Any],
    ) -> None:
        """Constructor method.

        Args:
            resource: An OpenAIResource instance.
            messages: A list of `Message` instances.
            *args: Positional arguments to pass to the parent constructor.
            **kwargs: Keyword arguments to pass to the parent constructor.
        """
        super().__init__(*args, **kwargs)

        self.resource = resource
        self.messages = messages

    async def stream_response(self, send: Send) -> None:
        """Stream chat completions.

        If an exception occurs while iterating over the OpenAI resource, an
        internal server error is sent to the client.

        Args:
            send: The ASGI send callable.
        """
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )

        try:
            async for chunk in self.resource.stream_response(self.messages):
                event_body = ServerSentEvent(
                    data=chunk,
                    event=Events.COMPLETION,
                )
                await send(
                    {
                        "type": "http.response.body",
                        "body": ensure_bytes(event_body, None),
                        "more_body": True,
                    }
                )
        except Exception as e:
            logger.error(f"openai error: {e}")
            error_event_body = ServerSentEvent(
                data=dict(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    detail=HTTPStatusDetail.INTERNAL_SERVER_ERROR,
                ),
                event=Events.ERROR,
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": ensure_bytes(error_event_body, None),
                    "more_body": True,
                }
            )

        await send({"type": "http.response.body", "body": b"", "more_body": False})



import sys
from typing import Any

import loguru


def get_logger(handler: Any = sys.stderr, **kwargs: dict[str, Any]) -> Any:
    """Get a logger instance.

    Lanarky uses `loguru` for logging.

    Args:
        handler: The handler to use for the logger.

    Returns:
        A loguru logger instance.
    """
    logger = loguru.logger
    logger.remove()
    logger.add(handler, **kwargs)
    return logger


logger = get_logger()



from typing import Any

import pydantic
from pydantic.fields import FieldInfo

try:
    from enum import StrEnum  # type: ignore
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum): ...


PYDANTIC_V2 = pydantic.VERSION.startswith("2.")


def model_dump(model: pydantic.BaseModel, **kwargs) -> dict[str, Any]:
    """Dump a pydantic model to a dictionary.

    Args:
        model: A pydantic model.
    """
    if PYDANTIC_V2:
        return model.model_dump(**kwargs)
    else:
        return model.dict(**kwargs)


def model_dump_json(model: pydantic.BaseModel, **kwargs) -> str:
    """Dump a pydantic model to a JSON string.

    Args:
        model: A pydantic model.
    """
    if PYDANTIC_V2:
        return model.model_dump_json(**kwargs)
    else:
        return model.json(**kwargs)


def model_fields(model: pydantic.BaseModel) -> dict[str, FieldInfo]:
    """Get the fields of a pydantic model.

    Args:
        model: A pydantic model.
    """
    if PYDANTIC_V2:
        return model.model_fields
    else:
        return model.__fields__
